from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist
import logging
from rest_framework import authentication
from rest_framework import exceptions


from ..util import normalize_auth_header


log = logging.getLogger(__name__)


class AuthTokenAuthentication(authentication.BaseAuthentication):
    def authenticate(self, request):
        log.debug('Fetching AuthToken header.')

        authz = authentication.get_authorization_header(request).split()
        if not authz or authz[0].lower() != b'authtoken':
            return None

        if len(authz) == 1:
            raise exceptions.AuthenticationFailed(
                'Invalid token header. No credentials provided.'
            )
        elif len(authz) > 2:
            raise exceptions.AuthenticationFailed(
                'Invalid token header. Token should not contain spaces.'
            )

        auth_type, data = authz
        email, auth_token = data.split(':', 1)

        log.debug('     email: %r', email)
        log.debug('auth_token: %r', auth_token)

        return self.authenticate_credentials(email, auth_token)

    def authenticate_credentials(self, email, auth_token):
        user = get_user_model().verify_auth_token(email, auth_token)

        # If user is bad this time, it's an invalid login
        if user is None:
            raise exceptions.AuthenticationFailed(
                'Invalid login/token expired.'
            )
            # raise exc.Unauthorized('Invalid login/token expired.')

        log.debug('token_auth authenticated user: %s' % email)

        return (user, auth_token)

    def authenticate_header(self, request):
        return 'AuthToken'


class EmailHeaderAuthentication(authentication.BaseAuthentication):
    def authenticate(self, request):
        user_auth_header = normalize_auth_header(settings.USER_AUTH_HEADER)
        log.debug('EmailHeaderAuthentication.authenticate(): auth_header = %r',
                  user_auth_header)

        # Naively fetch the user email from the auth_header
        email = request.META.get(user_auth_header)
        log.debug('EmailHeaderAuthentication.authenticate(): email = %r',
                  email)
        if email is None:
            return None

        # Fetch a stinkin' user
        try:
            user = get_user_model().objects.get(email=email)
        except ObjectDoesNotExist:
            # Make this a 400 for now since it's failing validation.
            raise exceptions.ValidationError(
                'Username must contain a valid email address'
            )

        # And return it.
        return (user, None)


class SecretKeyAuthentication(authentication.BaseAuthentication):
    def authenticate_credentials(self, email, secret_key):
        try:
            user = get_user_model().objects.get(email=email)
        except ObjectDoesNotExist:
            user = None

        # Make sure we've got a user and the secret_key is valid
        if user is not None and user.verify_secret_key(secret_key):
            return user, secret_key  # Auth success

        raise exceptions.AuthenticationFailed(
            'Invalid email/secret_key'
        )



from __future__ import unicode_literals

import ast
from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError as DjangoValidationError
from collections import OrderedDict
import json
import logging
from rest_framework import fields, serializers

from . import auth
from .. import exc
from .. import models


log = logging.getLogger(__name__)


class NsotSerializer(serializers.ModelSerializer):
    """Base serializer that logs change events."""
    def to_internal_value(self, data):
        """Inject site_pk from view's kwargs if it's not already in data."""
        view = self.context['view']
        kwargs = view.kwargs

        log.debug('NsotSerializer.to_internal_value() data [before] = %r', data)

        if 'site_id' not in data and 'site_pk' in kwargs:
            data['site_id'] = kwargs['site_pk']

        log.debug('NsotSerializer.to_internal_value() data [after] = %r', data)

        return super(NsotSerializer, self).to_internal_value(data)


######
# User
######
class UserSerializer(serializers.ModelSerializer):
    """
    UserProxy model serializer that takes optional `with_secret_key` argument
    that controls whether the secret_key for the user should be displayed.
    """
    def __init__(self, *args, **kwargs):
        # Don't pass `with_secret_key` up to the superclass
        with_secret_key = kwargs.pop('with_secret_key', None)
        super(UserSerializer, self).__init__(*args, **kwargs)

        # If we haven't passed `with_secret_key`, don't show the secret_key
        # field.
        if with_secret_key is None:
            self.fields.pop('secret_key')

    permissions = fields.ReadOnlyField(source='get_permissions')

    class Meta:
        model = get_user_model()
        fields = ('id', 'email', 'permissions', 'secret_key')

######
# Site
######
class SiteSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Site

    def to_representation(self, obj):
        return obj.to_dict()


#########
# Changes
#########
class ChangeSerializer(NsotSerializer):
    """Used for displaying Change events."""
    class Meta:
        model = models.Change

    def to_representation(self, obj):
        return obj.to_dict()


###########
# Attribute
###########
class AttributeSerializer(NsotSerializer):
    """Used for GET, DELETE on Attributes."""
    class Meta:
        model = models.Attribute

    def to_representation(self, obj):
        return obj.to_dict()


class JSONDictField(fields.Field):
    """Field used to represention attributes as JSON <-> Dict."""
    def to_representation(self, value):
        return value

    def to_internal_value(self, data):
        log.debug('JSONDictField.to_internal_value() data = %r', data)
        if not data:
            data = {}

        if isinstance(data, dict):
            return data

        # Try it as a regular JSON dict
        try:
            return json.loads(data)
        except ValueError:
            # Or try it as a Python dict
            try:
                return ast.literal_eval(data)
            except (SyntaxError, ValueError) as err:
                raise exc.ValidationError(err)
        except Exception as err:
            raise exc.ValidationError(err)
        return data


class AttributeCreateSerializer(AttributeSerializer):
    """Used for POST on Attributes."""
    constraints = JSONDictField(required=False)
    description = fields.CharField(allow_blank=True, required=False)
    site_id = fields.IntegerField()

    class Meta:
        model = models.Attribute
        fields = ('name', 'description', 'resource_name', 'required', 'display',
                  'multi', 'constraints', 'site_id')


class AttributeUpdateSerializer(AttributeCreateSerializer):
    """Used for PUT, PATCH, on Attributes."""
    class Meta:
        model = models.Attribute
        fields = ('description', 'required', 'display', 'multi', 'constraints')


#######
# Value
#######
class ValueSerializer(serializers.ModelSerializer):
    """Used for GET, DELETE on Values."""
    class Meta:
        model = models.Value
        fields = ('id', 'name', 'value', 'attribute', 'resource_name',
                  'resource')

    # Not sure if we want to view an attribute value w/ so much context just
    # yet.
    # def to_representation(self, obj):
    #      return obj.to_dict()


class ValueCreateSerializer(ValueSerializer):
    """Used for POST on Values."""
    class Meta:
        model = models.Value
        read_only_fields = ('id', 'name', 'resource_name')
        fields = ('id', 'name', 'value', 'attribute', 'resource_name',
                  'resource')


########
# Device
########
class DeviceSerializer(NsotSerializer):
    """Used for GET, DELETE on Devices."""
    class Meta:
        model = models.Device

    def to_representation(self, obj):
        if isinstance(obj, OrderedDict):
            return obj
        return obj.to_dict()


class DeviceCreateSerializer(DeviceSerializer):
    """Used for POST on Devices."""
    attributes = JSONDictField(required=False)
    site_id = fields.IntegerField()

    class Meta:
        model = models.Device
        fields = ('hostname', 'attributes', 'site_id')

    def create(self, validated_data):
        attributes = validated_data.pop('attributes', {})
        obj = super(DeviceCreateSerializer, self).create(validated_data)
        try:
            obj.set_attributes(attributes)
        except exc.ValidationError:
            obj.delete()
            raise
        obj.save()
        return obj


class DeviceUpdateSerializer(DeviceCreateSerializer):
    """Used for PUT, PATCH, on Devices."""
    class Meta:
        model = models.Device
        fields = ('hostname', 'attributes')

    def update(self, instance, validated_data):
        attributes = validated_data.pop('attributes', {})
        obj = super(DeviceUpdateSerializer, self).update(
            instance, validated_data
        )
        obj.set_attributes(attributes)
        obj.save()
        return obj


#########
# Network
#########
class NetworkSerializer(NsotSerializer):
    """Used for GET, DELETE on Networks."""
    class Meta:
        model = models.Network

    def to_representation(self, obj):
        return obj.to_dict()


class NetworkCreateSerializer(NetworkSerializer):
    """Used for POST on Networks."""
    attributes = JSONDictField(required=False)
    cidr = fields.CharField(write_only=True)
    site_id = fields.IntegerField()

    class Meta:
        model = models.Network
        fields = ('cidr', 'attributes', 'site_id')

    def create(self, validated_data):
        attributes = validated_data.pop('attributes', {})
        obj = super(NetworkCreateSerializer, self).create(validated_data)
        try:
            obj.set_attributes(attributes)
        except exc.ValidationError:
            obj.delete()
            raise
        obj.save()
        return obj


class NetworkUpdateSerializer(NetworkCreateSerializer):
    """Used for PUT, PATCH on Networks."""
    class Meta:
        model = models.Network
        fields = ('attributes',)

    def update(self, instance, validated_data):
        log.debug('NetworkUpdateSerializer.update() validated_data = %r',
                  validated_data)
        attributes = validated_data.pop('attributes', {})
        obj = super(NetworkUpdateSerializer, self).update(
            instance, validated_data
        )
        obj.set_attributes(attributes)
        obj.save()
        return obj


###########
# AuthToken
###########
class AuthTokenSerializer(serializers.Serializer):
    """
    AuthToken authentication serializer to validate username/secret_key inputs.
    """
    email = serializers.CharField()
    secret_key = serializers.CharField()

    def validate(self, attrs):
        email = attrs.get('email')
        secret_key = attrs.get('secret_key')

        if email and secret_key:
            auth_func = auth.SecretKeyAuthentication().authenticate_credentials
            user, secret_key = auth_func(email, secret_key)

            if user:
                if not user.is_active:
                    msg = 'User account is disabled.'
                    raise exc.ValidationError(msg)
                attrs['user'] = user
                return attrs
            else:
                msg = 'Unable to login with provided credentials.'
                raise exc.ValidationError(msg)
        else:
            msg = 'Must include "email" and "secret_key"'
            raise exc.ValidationError(msg)



from __future__ import absolute_import, print_function

import sys

from django.core.management import call_command
from django.core.management.base import BaseCommand, CommandError
import logging
from optparse import make_option


class Command(BaseCommand):
    args = '<service>'
    help = 'Starts the specified service'

    option_list = BaseCommand.option_list + (
        make_option('--debug',
            action='store_true',
            dest='debug',
            default=False),
        make_option('--noupgrade',
            action='store_false',
            dest='upgrade',
            default=True),
        make_option('--workers', '-w',
            dest='workers',
            type=int,
            default=None),
        make_option('--worker-class', '-k',
            dest='worker_class',
            type=str,
            default=None),
        make_option('--noinput',
            action='store_true',
            dest='noinput',
            default=False,
            help='Tells Django to NOT prompt the user for input of any kind.',
        ),
    )

    def handle(self, service_name='http', address=None, upgrade=True, **options):
        from nsot.services import http

        if address:
            if ':' in address:
                host, port = address.split(':', 1)
                port = int(port)
            else:
                host = address
                port = None
        else:
            host, port = None, None

        services = {
            'http': http.NsotHTTPServer,
        }

        if upgrade:
            # Ensure we perform an upgrade before starting any service
            print("Performing upgrade before service startup...")
            call_command('upgrade', verbosity=0, noinput=options.get('noinput'))

        try:
            service_class = services[service_name]
        except KeyError:
            raise CommandError('%r is not a valid service' % service_name)

        service = service_class(
            debug=options.get('debug'),
            host=host,
            port=port,
            workers=options.get('workers'),
            worker_class=options.get('worker_class')
        )

        # Remove command line arguments to avoid optparse failures with service
        # code that calls call_command which reparses the command line, and if
        # --noupgrade is supplied a parse error is thrown.
        sys.argv = sys.argv[:1]

        print('Running service: %r' % service_name)
        service.run()



from __future__ import absolute_import, print_function

from gunicorn.app.base import Application

from nsot.services.base import Service


class NsotGunicornCommand(Application):
    def __init__(self, options):
        self.usage = None
        self.prog = None
        self.cfg = None
        self.config_file = ""
        self.options = options
        self.callable = None
        self.project_path = None
        self.do_load_config()

    def init(self, *args):
        cfg = {}
        for k, v in self.options.items():
            if k.lower() in self.cfg.settings and v is not None:
                cfg[k.lower()] = v
        return cfg

    def load(self):
        import nsot.wsgi
        return nsot.wsgi.application


class NsotHTTPServer(Service):
    name = 'http'

    def __init__(self, host=None, port=None, debug=False, workers=None,
                 worker_class=None):
        from django.conf import settings

        self.host = host or settings.NSOT_HOST
        self.port = port or settings.NSOT_PORT
        self.workers = workers

        options = {}
        options.setdefault('bind', '%s:%s' % (self.host, self.port))
        options.setdefault('timeout', 30)
        options.setdefault('proc_name', 'NSoT')
        options.setdefault('workers', 4)
        options.setdefault('worker_class', 'gevent')
        options.setdefault('access_logfile', '-')
        options.setdefault('errorlog', '-')
        options.setdefault('loglevel', 'info')
        options.setdefault('limit_request_line', 0)
        options['preload'] = False

        if workers:
            options['workers'] = workers

        if worker_class:
            options['worker_class'] = worker_class

        self.options = options

    def run(self):
        NsotGunicornCommand(self.options).run()



import pytest
from pytest_django.fixtures import  django_user_model, transactional_db
import logging

from nsot import models


log = logging.getLogger(__name__)


@pytest.fixture
def user(django_user_model):
    """Create and return a non-admin user."""
    user = django_user_model.objects.create(email='user@localhost')
    return user


@pytest.fixture
def admin_user(django_user_model):
    """Create and return an admin user."""
    user = django_user_model.objects.create(
        email='admin@localhost', is_superuser=True, is_staff=True
    )
    return user


@pytest.fixture
def site():
    site = models.Site.objects.create(
        name='Test Site', description='This is a Test Site.'
    )
    return site
